#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""plugin: command. Used to execute a user defined command or script"""

import logging
import os
import shutil
import subprocess
from typing import List

import click
from c8ylp import options

from c8ylp.cli.core import ExitCodes, ProxyContext


@click.command()
@options.common_options
@click.argument(
    "additional_args",
    metavar="[REMOTE_COMMANDS]...",
    nargs=-1,
    type=click.UNPROCESSED,
)
@click.pass_context
def cli(ctx: click.Context, additional_args: List[str], **kwargs):
    """
    Start once-off proxy and execute a (local) script/command

        \b
        DEVICE is the device's external identity
        REMOTE_COMMANDS is the script or command to run after the proxy has been started

        All additional arguments will be passed to the script/command. Use "--" before
        the additional arguments to prevent the arguments from being interpreted by
        c8ylp (i.e. to avoid clashes with c8ylp).

    \b
    Available ENV variables (use single quotes to prevent expansion):

      \b
      DEVICE - external device identity
      PORT   - port number of the local proxy
      C8Y_HOST - Cumulocity host/url

    \b
    Example 1: Use scp to copy a file to a device

        \b
        c8ylp plugin command --env-file .env device01 \\
            -- /usr/bin/scp -P '$PORT' myfile.tar.gz admin@localhost:/tmp

    Example 2: Run a custom script (not included) to copy a file from the device to
    the current folder

        \b
        c8ylp plugin command --env-file .env -v device01 ./copyfrom.sh /var/log/dpkg.log ./
    """

    cmd_args = []

    if not additional_args:
        raise click.BadParameter("At least one argument needs to be provided")

    proxy = ProxyContext(ctx, kwargs).start_background()

    if additional_args:
        for value in additional_args:
            expanded_value = os.path.expandvars(value)
            logging.info("Expanded script arguments: %s => %s", value, expanded_value)
            cmd_args.append(expanded_value)

    logging.info("PATH: %s", os.getenv("PATH"))
    first_arg = shutil.which(cmd_args[0])

    if first_arg:
        # resolve path to its full form to prevent other execution errors
        # on windows
        cmd_args[0] = first_arg
    else:
        # Prepend a bash, as the user is probably trying to use
        # a bash function (like echo)
        bash_path = shutil.which("bash")
        if not bash_path:
            proxy.show_error(f"Command does not exist. cmd={cmd_args}")
            ctx.exit(ExitCodes.COMMAND_NOT_FOUND)

        proxy.show_info("Using bash to execute the command")
        cmd_args = [bash_path, "-c"] + [" ".join(cmd_args)]

    proxy.show_message(f"Running custom command: {' '.join(cmd_args)}")

    exit_code = subprocess.call(cmd_args, env=os.environ, shell=False)
    if exit_code != 0:
        proxy.show_error(f"Command exited with a non-zero exit code. code={exit_code}")
    ctx.exit(exit_code)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test plugin command"""

import os
import subprocess
import sys

import pytest

from .fixtures import Device


def proxy_cli(*args, **kwargs) -> subprocess.CompletedProcess:
    """Execute the proxy cli command with given arguments"""
    return subprocess.run(
        [
            sys.executable,
            "-m",
            "c8ylp",
            *args,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )


@pytest.mark.parametrize(
    "case",
    (
        dict(command="sleep 1s", exit_code=0),
        dict(command="sleep 1s; exit 111", exit_code=111),
        dict(command="sleep 1s; exit 111", user="unknown_user", exit_code=255),
    ),
    ids=lambda x: str(x),
)
def test_ssh_command_then_exit(case, c8ydevice: Device):
    """Test running a once off ssh command"""
    user = case.get("user", c8ydevice.ssh_user)
    command = case.get("command", "sleep 10s")

    result = proxy_cli(
        "connect",
        "ssh",
        c8ydevice.device,
        "--ssh-user",
        user,
        "--",
        command,
    )

    assert result.returncode == case.get("exit_code", 0)


@pytest.mark.parametrize(
    "case",
    (dict(command="sleep 1s", exit_code=0),),
    ids=lambda x: str(x),
)
def test_support_ssh_user_via_env(case, c8ydevice: Device):
    """Test ssh command should support setting the ssh user via env variable"""
    user = case.get("user", c8ydevice.ssh_user)
    command = case.get("command", "sleep 10s")

    result = proxy_cli(
        "connect",
        "ssh",
        c8ydevice.device,
        "--",
        command,
        env={
            **os.environ,
            "C8YLP_SSH_USER": user,
        },
    )

    assert result.returncode == case.get("exit_code", 0)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Options"""

import logging
import os
from typing import Any
import functools

import click

from c8ylp.env import loadenv
from c8ylp.rest_client.c8yclient import CumulocityClient


def load_envfile(ctx: click.Context, _param: click.Parameter, value: Any):
    """Load environment variables from a file

    Args:
        ctx (click.Context): Click context
        _param (click.Parameter): Click parameter
        value (Any): Parameter value
    """
    if not value or ctx.resilient_parsing:
        return value

    click.echo(f"Loading env-file: {value}")
    if os.path.exists(value) and os.path.isfile(value):
        loadenv(value)
    else:
        logging.info("env file does not exist or is not a file: %s", value)
    return value


def deactivate_prompts(ctx: click.Context, _param: click.Parameter, value: Any):
    """Deactivate all prompts

    Args:
        ctx (click.Context): Click Context
        _param (click.Parameter): Click Parameter
        value (Any): Parameter value

    Returns:
        Any: Parameter value
    """
    if value:
        for i_param in ctx.command.params:
            if isinstance(i_param, click.Option) and i_param.prompt is not None:
                i_param.prompt = None
    return value


def validate_token(ctx: click.Context, _param, value) -> Any:
    """Validate Cumulocity token. If it is invalid then it
    will be ignored.

    Args:
        ctx (Any): Click context
        _param (Any): Click param
        value (Any): Parameter value

    Returns:
        Any: Parameter value
    """
    if not value or ctx.resilient_parsing:
        return None

    host = ctx.params.get("host")

    if not host:
        return value

    if not value:
        return value

    if isinstance(value, tuple):
        return value

    client = CumulocityClient(hostname=host, token=value)

    try:
        client.validate_credentials()
        click.secho("Validating c8y token: ", nl=False)
        click.secho("OK", fg="green")
    except Exception:
        click.secho("Validating c8y token: ", nl=False)
        click.secho("EXPIRED/INVALID", fg="red")
        logging.info(
            "Token is no longer valid for host. The token will be ignored. host=%s",
            host,
        )
        return ""
    return value


def trim_trailing_slash(ctx: click.Context, _param, value) -> Any:
    """Trim any trailing slashes

    Args:
        ctx (Any): Click context
        _param (Any): Click param
        value (Any): Parameter value

    Returns:
        Any: Parameter value
    """
    if not value or ctx.resilient_parsing:
        return None

    # Trim both space then trailing slashes
    return str(value).strip().rstrip("/")


def deprecated(ctx: click.Context, _param, value) -> Any:
    """Deprecated option warning

    Args:
        ctx (Any): Click context
        _param (Any): Click param
        value (Any): Parameter value

    Returns:
        Any: Parameter value
    """
    name = getattr(_param, "name", "")
    if name and value is not None:
        click.secho(
            f"Warning: '{name}' option is deprecated. It will be removed in the next major release",
            fg="yellow",
        )
    if not value or ctx.resilient_parsing:
        return None

    return value


HOSTNAME = click.option(
    "--host",
    "host",
    is_eager=True,
    prompt=False,
    callback=trim_trailing_slash,
    envvar=("C8Y_HOST", "C8Y_BASEURL", "C8Y_URL"),
    help="Cumulocity Hostname  [required] [env var: C8Y_HOST]",
)

HOSTNAME_PROMPT = click.option(
    "--host",
    "host",
    is_eager=False,
    prompt=True,
    callback=trim_trailing_slash,
    envvar=("C8Y_HOST", "C8Y_BASEURL", "C8Y_URL"),
    help="Cumulocity Hostname  [required] [env var: C8Y_HOST]",
)


ARG_DEVICE = click.argument(
    "device",
    nargs=1,
    required=True,
)

DEVICE = click.option(
    "--device",
    "-d",
    required=False,
    envvar="C8YLP_DEVICE",
    show_envvar=True,
    help="Device external identity",
)

EXTERNAL_IDENTITY_TYPE = click.option(
    "--external-type",
    envvar="C8YLP_EXTERNAL_TYPE",
    default="c8y_Serial",
    show_default=True,
    show_envvar=True,
    help="external Id Type",
)

REMOTE_ACCESS_TYPE = click.option(
    "--config",
    "-c",
    required=False,
    envvar="C8YLP_CONFIG",
    default="Passthrough",
    show_default=True,
    show_envvar=True,
    help="name of the C8Y Remote Access Configuration",
)

C8Y_TENANT = click.option(
    "--tenant", "-t", envvar="C8Y_TENANT", help="Cumulocity tenant id", show_envvar=True
)

C8Y_USER = click.option(
    "--user",
    "-u",
    envvar=("C8Y_USER", "C8Y_USERNAME"),
    show_envvar=True,
    help="Cumulocity username",
)

C8Y_TOKEN = click.option(
    "--token",
    callback=validate_token,
    envvar="C8Y_TOKEN",
    is_eager=True,
    show_envvar=True,
    help="Cumulocity token",
)

C8Y_PASSWORD = click.option(
    "--password",
    "-p",
    envvar="C8Y_PASSWORD",
    prompt=False,
    hide_input=True,
    show_envvar=True,
    help="Cumulocity password",
)

C8Y_TFA_CODE = click.option(
    "--tfa-code",
    envvar="C8Y_TFA_CODE",
    show_envvar=True,
    help="TFA Code. Required when the 'TFA enabled' is enabled for a user",
)

PORT = click.option(
    "--port",
    type=click.IntRange(0, 65535),
    default=2222,
    envvar="C8YLP_PORT",
    show_envvar=True,
    show_default=True,
    help="TCP Port which should be opened. 0=Random port",
)

PORT_DEFAULT_RANDOM = click.option(
    "--port",
    type=click.IntRange(0, 65535),
    default=0,
    envvar="C8YLP_PORT",
    show_default=True,
    show_envvar=True,
    help="TCP Port which should be opened. 0=Random port",
)

PING_INTERVAL = click.option(
    "--ping-interval",
    type=int,
    default=0,
    show_default=True,
    show_envvar=True,
    envvar="C8YLP_PING_INTERVAL",
    help="Websocket ping interval in seconds. 0=disabled",
)

TCP_SIZE = click.option(
    "--tcp-size",
    envvar="C8YLP_TCP_SIZE",
    type=click.IntRange(1024, 8096 * 1024),
    default=4096,
    show_default=True,
    show_envvar=True,
    help="TCP Package Size",
)

TCP_TIMEOUT = click.option(
    "--tcp-timeout",
    envvar="C8YLP_TCP_TIMEOUT",
    default=0,
    show_default=True,
    show_envvar=True,
    help="Timeout in sec. for inactivity. Can be activated with values > 0",
)

LOGGING_VERBOSE = click.option(
    "--verbose",
    "-v",
    envvar="C8YLP_VERBOSE",
    is_flag=True,
    default=False,
    show_envvar=True,
    help="Print Debug Information into the Logs and Console when set",
)

STORE_TOKEN = click.option(
    "--store-token/--no-store-token",
    "store_token",
    envvar="C8YLP_STORE_TOKEN",
    is_flag=True,
    default=True,
    show_envvar=True,
    help="Store the Cumulocity host, tenant and token to the env-file if a file is being used",
)


DISABLE_PROMPT = click.option(
    "--disable-prompts",
    "-d",
    "disable_prompts",
    envvar="C8YLP_DISABLE_PROMPTS",
    default=False,
    is_eager=True,
    is_flag=True,
    expose_value=True,
    show_envvar=True,
    callback=deactivate_prompts,
)

SSL_IGNORE_VERIFY = click.option(
    "--ignore-ssl-validate",
    envvar="C8YLP_IGNORE_SSL_VALIDATE",
    is_flag=True,
    default=False,
    show_envvar=True,
    help="Ignore Validation for SSL Certificates while connecting to Websocket",
)


SERVER_RECONNECT_LIMIT = click.option(
    "--reconnects",
    envvar="C8YLP_RECONNECTS",
    type=str,
    default=None,
    show_default=True,
    show_envvar=True,
    hidden=True,
    callback=deprecated,
    help="[Deprecated] No longer used but kept in for compatibility until next major release",
)

ARG_SSH_COMMAND = click.argument("command", nargs=1, required=True)

ARG_SCRIPT = click.argument(
    "script", type=click.Path(resolve_path=True), nargs=1, required=True
)


ENV_FILE = click.option(
    "--env-file",
    "env_file",
    envvar="C8YLP_ENV_FILE",
    is_eager=True,
    expose_value=True,
    show_envvar=True,
    type=click.Path(
        exists=True,
    ),
    callback=load_envfile,
    help="Environment file to load. Any settings loaded via this file will control other options",
)

ENV_FILE_OPTIONAL_EXISTS = click.option(
    "--env-file",
    "env_file",
    envvar="C8YLP_ENV_FILE",
    is_eager=True,
    expose_value=True,
    show_envvar=True,
    required=False,
    type=click.Path(
        exists=False,
    ),
    callback=load_envfile,
    help="Environment file to load. Any settings loaded via this file will control other options",
)


def common_options(f):
    """Common Options"""
    options = [
        ARG_DEVICE,
        HOSTNAME,
        C8Y_TENANT,
        C8Y_USER,
        C8Y_TOKEN,
        C8Y_PASSWORD,
        C8Y_TFA_CODE,
        ENV_FILE,
        EXTERNAL_IDENTITY_TYPE,
        REMOTE_ACCESS_TYPE,
        PORT_DEFAULT_RANDOM,
        PING_INTERVAL,
        TCP_SIZE,
        TCP_TIMEOUT,
        LOGGING_VERBOSE,
        SSL_IGNORE_VERIFY,
        STORE_TOKEN,
        DISABLE_PROMPT,
        SERVER_RECONNECT_LIMIT,
    ]

    # Need to reverse the order to control the list order
    options = reversed(options)
    return functools.reduce(lambda x, opt: opt(x), options, f)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Environment utilities"""

import os
import pathlib
from collections import defaultdict
from typing import Any, Dict


def loadenv(path: str) -> None:
    """Load environment variables from file

    Lines are ignored if they match any of the following
     * Empty line (only whitespace)
     * Lines starts with "#"
     * Lines starts with ";"

    Values may contain references to other environment variables by
    using "$NAME" or "${NAME}" syntax.

    Values with single or double quotes will be stripped from the value.

    Below is an example of a dotenv file:
    >>>
    ENV1=MYVALUE
    CUSTOMER_NAME="$USER"

    # Literal values can be surrounded with single quotes.
    CUSTOMER_NAME='$USER'
    >>>

    Args:
        path (str): Dotenv file path
    """
    with open(path, "r", encoding="utf8") as file:
        for line in file:
            line = line.strip()
            if not line or line.startswith("#") or line.startswith(";"):
                continue

            key, _, value = line.partition("=")

            if value.startswith("'") and value.endswith("'"):
                # Strip single quote wrapping, and skip env expansion
                value = value.strip("'")
            elif value.startswith('"') and value.endswith('"'):
                # Strip surrounding quotes
                value = os.path.expandvars(value.strip('"'))
            else:
                value = os.path.expandvars(value)

            os.environ[key] = value


def save_env(path: str, values: Dict[str, Any]) -> bool:
    """Save a new value to an environment file. If the key already
    exists, then it will be updated in place, otherwise it will be appended.

    If the file does not exist, then it will be created.

    The order of the environment variables are preserved

    Args:
        path (str): file path
        values (Dict[str, Any]): New key/value pairs to add to env file

    Returns:
        bool: True if the file has been changed
    """
    env_file = pathlib.Path(path)
    has_changed = False
    output = (
        []
        if not env_file.exists()
        else env_file.read_text(encoding="utf8").splitlines()
    )

    key_indexes = defaultdict(lambda: [])

    # collect line indexes of key in existing env file
    for i, line in enumerate(output):
        key, _, _ = line.partition("=")
        if key:
            key_indexes[key].append(i)

    # updates values (if changed, and keep order)
    for key, value in values.items():
        d_value = value if value is not None else ""
        newline = f"{key}={d_value}"

        if key in key_indexes:
            # existing key (update if changed)
            for existing_index in key_indexes[key]:
                if output[existing_index] != newline:
                    output[existing_index] = newline
                    has_changed = True
        else:
            # new key
            output.append(newline)
            has_changed = True

    if has_changed:
        env_file.write_text("\n".join(output).rstrip("\n") + "\n", encoding="utf8")

    return has_changed



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Server tests"""

from unittest.mock import patch

from click.testing import CliRunner
from c8ylp.main import cli


@patch("c8ylp.cli.server.ProxyContext", autospec=True)
def test_server_mode(mock_context):
    """Start a mocked server"""
    mock_context.start.return_value = 0

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["server", "ext-device-01"],
    )
    assert result.exit_code == 0



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test sessions"""

import responses
from c8ylp.rest_client.sessions import BaseUrlSession


@responses.activate
def test_session_with_partial_urls():
    """Test expansion of partial urls when sending requests"""
    session = BaseUrlSession("example.c8y.io")

    responses.add(
        "GET",
        url="https://example.c8y.io/partial/url",
    )
    response = session.get("/partial/url")
    assert response.status_code == 200
    assert response.url == "https://example.c8y.io/partial/url"


@responses.activate
def test_session_with_full_url():
    """Test handling of full urls. The hostname should be left untouched"""
    session = BaseUrlSession("example.c8y.io")

    responses.add(
        "GET",
        url="https://someotherurl.com/hello",
    )
    response = session.get("https://someotherurl.com/hello")
    assert response.status_code == 200
    assert response.url == "https://someotherurl.com/hello"



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Help tests"""
import re

from click.testing import CliRunner
from c8ylp.main import cli
from tests.env import Environment


def test_show_version(env: Environment):
    """Test show version"""
    runner = CliRunner()
    result = runner.invoke(cli, ["version"], env=env.create_empty_env())
    assert re.match(r"Version \S+\n", result.output) is not None



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tasks"""
import os
import re
import shutil
import subprocess
import sys
from invoke import task
from pathlib import Path


@task
def clean(c, docs=False, bytecode=False, extra=""):
    """Clean project"""
    patterns = ["build", "dist", "test_output", "deb_dist", "c8ylp*tar.gz"]
    if docs:
        patterns.append("docs/cli")
    if bytecode:
        patterns.append("**/*.pyc")
    if extra:
        patterns.append(extra)
    for pattern in patterns:
        for match in Path(".").glob(pattern):
            if match.is_file():
                match.unlink(missing_ok=True)
            else:
                shutil.rmtree(match)


@task
def lint(c):
    """Lint"""
    c.run("pylint c8ylp")
    c.run("pylint tests/")


@task
def format(c, check=False):
    """Format code (using black)"""
    if check:
        c.run("black --check --target-version=py37 .")
    else:
        c.run(f"{sys.executable} -m black --target-version=py37 .")


@task
def build(c):
    """Build"""
    c.run(f"{sys.executable} setup.py sdist bdist_wheel")


@task(pre=[clean])
def build_deb(c):
    """Build debian package - (Debian/Ubuntu only)

    apt-get install -y python-all dh-python python3-stdeb
    """
    print("Warning: Requires package 'python-all dh-python python3-stdeb' to work")
    c.run(f"{sys.executable} setup.py --command-packages=stdeb.command bdist_deb")

    print("\nThe debian file is stored under: deb_dist/*.deb\n")


@task(pre=[build])
def publish(c):
    """Publish python package to pip"""
    if "TWINE_USERNAME" not in os.environ or "TWINE_PASSWORD" not in os.environ:
        print(
            "'TWINE_USERNAME' and 'TWINE_PASSWORD' environment variables "
            "are required for publishing"
        )
        sys.exit(1)

    c.run("twine upload dist/*")


@task
def test(c, pattern=None):
    """Run unit tests and coverage report"""
    cmd = [
        "pytest",
        "tests",
        "--timeout=10",
        # Note: Don't use log cli level (--log-cli-level) as it can affect click testing!
        "--cov-config=.coveragerc",
        "--cov-report=term",
        "--cov-report=html:test_output/htmlcov",
        "--cov=c8ylp",
    ]

    if pattern:
        cmd.append(f"-k={pattern}")
    c.run(" ".join(cmd))


@task(pre=[clean])
def test_integration(c, pattern=None):
    """Run integration tests"""
    cmd = [
        sys.executable,
        "-m",
        "pytest",
        "--durations=0",
        "--timeout=3600",
        "--log-cli-level=INFO",
        "--cov-config=.coveragerc",
        "--cov-report=term",
        "--cov-report=html:test_output/htmlcov",
        "--cov=c8ylp",
    ]

    assert os.path.exists(".env") or os.environ.get(
        "C8Y_HOST"
    ), "Missing Cumulocity configuration required for integration tests"

    if pattern:
        cmd.append(f"-k={pattern}")
    c.run(" ".join(cmd))


@task
def generate_docs(c):
    """Generate cli docs (markdown files)"""
    commands = [
        ("c8ylp",),
        (
            "c8ylp",
            "login",
        ),
        (
            "c8ylp",
            "server",
        ),
        (
            "c8ylp",
            "connect",
        ),
        (
            "c8ylp",
            "connect",
            "ssh",
        ),
        (
            "c8ylp",
            "plugin",
        ),
        (
            "c8ylp",
            "plugin",
            "command",
        ),
        (
            "c8ylp",
            "version",
        ),
    ]

    doc_dir = Path("docs") / "cli"
    doc_dir.mkdir(parents=True, exist_ok=True)

    for cmd in commands:
        name = "_".join(cmd).upper() + ".md"
        doc_file = doc_dir / name
        print(f"Updating cli doc: {str(doc_file)}")

        proc = subprocess.run(
            [sys.executable, "-m", *cmd, "--help"],
            stdout=subprocess.PIPE,
            env={
                **os.environ,
                "C8Y_HOST": "dummy",
                "C8YLP_PLUGINS": "dummy",
            },
        )

        usage = proc.stdout.decode()

        # Replace python3 module usage, to the more common c8ylp
        usage = re.sub("python3? -m ", "", usage)

        usage = re.sub(r"Validating c8y token: EXPIRED/INVALID.*\n", "", usage)

        # Strip plugin folder information out of usage
        usage = re.sub(r"Checking plugin folder.+\n", "", usage)

        doc_template = f"""
## {" ".join(cmd)}

```
{usage}
```
"""

        doc_file.write_text(doc_template)


@task
def add_license(c):
    """Add license header to source files"""
    c.run(
        f"{sys.executable} -m licenseheaders --current-year --owner 'Software AG, Darmstadt, Germany' -E .py -t licence_header.tmpl -d ."
    )
